"""Fixpoint iterator over a CfG

- Transfer is a base class that must be instantiated for
each static analysis.
- fixpoint_iteration computes a fixpoint over a cfg for the given
transfer instance
"""

from cfg import *
from typing import Protocol


class Transfer[S](Protocol):
    """lists the functions that must be implemented for a static analysis over Lattice S"""

    def bottom(self) -> S:
        """the empty state, associated to unvisited nodes"""
        ...

    def init_state(self) -> S:
        """the state associated to the initial node"""
        ...

    def join(self, s1: S, s2: S) -> S:
        """join of two states when merging branches"""
        ...

    def included(self, s1: S, s2: S) -> bool:
        """return True when the first state is included in the second"""
        ...

    def tr_skip(self, s: S) -> S:
        """transfer state across a Skip transition

        Almost always transfer the state unchanged
        """
        return s

    def tr_set(self, s: S, v: str, e: ArithExpr) -> S:
        """transfer state across an assignment"""
        ...

    def tr_test(self, s: S, c: BoolExpr) -> S:
        """transfer state across a test transition"""
        ...

    def tr_err(self, s: S, e: Expr) -> S:
        """transfer state across a transition to error state"""
        ...


def fixpoint_iteration[T](transfer: Transfer, cfg: Cfg) -> dict[Node, T]:
    """computes the fixpoint for the given transfer functions over the given CfG

    output maps each node to the computed state
    """
    mapping: dict[Node, T] = {}
    for node in cfg.g.nodes:
        if node == cfg.init_node:
            mapping[node] = transfer.init_state()
        else:
            mapping[node] = transfer.bottom()

    worklist = [(cfg.init_node, dest, lab) for (dest, lab) in cfg.init_node.succs]

    while worklist:
        src, dst, label = worklist.pop(0)
        s_src = mapping[src]
        if isinstance(label, LSkip):
            s_new = transfer.tr_skip(s_src)
        elif isinstance(label, LSet):
            s_new = transfer.tr_set(s_src, label.var, label.expr)
        elif isinstance(label, LTest):
            s_new = transfer.tr_test(s_src, label.cond)
        elif isinstance(label, LErr):
            s_new = transfer.tr_err(s_src, label.err)
        else:
            s_new = s_src
        new_state = transfer.join(mapping[dst], s_new)
        if not transfer.included(mapping[dst], new_state):
            mapping[dst] = new_state
            for succ, lab in dst.succs:
                worklist.append((dst, succ, lab))
    return mapping


"""
# À quoi correspondent les nœuds indiqués comme inatteignables par l'analyse?

Les nœuds inatteignables correspondent aux branches du programme qui ne peuvent jamais être exécutées, car leur condition s’évalue toujours à False dans l’analyse.
"""
